/*
ztensound.c : background sound playback module for zten(v)

Copyright (c) 2005 Takashi Nemoto

Permission is hereby granted, free of charge, to any person 
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without 
restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the 
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be 
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef MAC
#include "ztensound.h"

#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/soundcard.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>

#include <cstring>
#include <cstdio>
#define ZTENSOUNDLOCK "/tmp/ztensnd"

inline unsigned int getDWord(const unsigned char *ubuf) {
  return 256*256*256*ubuf[3]+256*256*ubuf[2]+256*ubuf[1]+ubuf[0];
}

inline unsigned short getWord(const unsigned char *ubuf) {
  return 256*ubuf[1]+ubuf[0];
}

inline unsigned int getDWord(const char *buf) {
  return getDWord((const unsigned char *) buf);
}

inline unsigned short getWord(const char *buf) {
  return getWord((const unsigned char *) buf);
}

inline int getSWord(const unsigned char *buf) {
  int s=getWord(buf);
  if (s>32767) return s-65536; else return s;
}

const unsigned char* decodeMSADPCM(const char *buf, int& size);

static pid_t pid=0;

// Main Process Side 
void StopSound() {
  if (pid>0) {
    int st;
    kill(pid,SIGKILL);
    usleep(1000);
    waitpid(pid, &st, WNOHANG);
    pid=0;
  }
}

bool PlaySound(const char *buf, int size) {
  const unsigned char *ubuf=(const unsigned char *)buf;
  int speed=44100, channel=1, wav_format, format=AFMT_S16_LE, dsp;
  bool delflag=false;

  if (pid!=0) {
    StopSound();
  }
  pid=fork();
  if (pid!=0) {
    return true;
  }

  if (size>=44 &&
      strncmp(buf,"RIFF",4)==0 &&
      strncmp(buf+8,"WAVE",4)==0 &&
      strncmp(buf+12,"fmt",3)==0) {
    // Tenuki....
    wav_format=getWord(buf+20);
    channel=getWord(buf+22);
    speed=getDWord(ubuf+24);
    if (ubuf[34]==8) {
      format=AFMT_U8;
    } else if (ubuf[34]==16) {
      format=AFMT_S16_LE;
    }
    if (wav_format==2) {
      format=AFMT_U8;
      ubuf = decodeMSADPCM(buf,size);
      buf=(const char *) ubuf;
      delflag=true;
    } else if (wav_format==0 || wav_format==1) {
      buf+=44;
      size-=44;
    }
  }

  unlink(ZTENSOUNDLOCK);

  int count=0;
  struct stat fst;


  size-=1; 

  dsp = open("/dev/dsp", O_WRONLY);
  if (dsp<0) goto ret_false;
  if (ioctl(dsp, SNDCTL_DSP_SETFMT, &format)<0) 
    fprintf(stderr,"SND:IOCTL:SETFMT error %d\n",format);
  if (ioctl(dsp, SNDCTL_DSP_CHANNELS, &channel)<0)
    fprintf(stderr,"SND:IOCTL:CHANNEL error %d\n",channel);
  if (ioctl(dsp, SNDCTL_DSP_SPEED, &speed)<0)
    fprintf(stderr,"SND:IOCTL:SPEED error %d\n",speed);

  while(size>0) {
    int len=8192;
    if (len>size) len=size;
    int len2=write(dsp, buf, len);
    if (len2<=0) break;
    size-=len2;
    buf+=len2;
    count++;
    if ((count&7)==0) {
      if (lstat(ZTENSOUNDLOCK,&fst)==0) break;
    }
  }

  char buf2[8192];
  for(int i=0;i<8192;i++)buf2[i]=0x7f;
  write(dsp, buf2,8192);

  close(dsp);

  // ret_true:
  // if (delflag && ubuf!=NULL) delete [] ubuf;
  //  return true;
  //  _exit(0);

 ret_false:
  if (delflag && ubuf!=NULL) delete [] ubuf;
  _exit(0);
  //  return false;
}

// MS ADPCM
// Reference : http://www.multimedia.cx/simpleaudio.html
const unsigned char* decodeMSADPCM(const char *sbuf, int& size) {
  unsigned char *obuf = new unsigned char [size*2];
  const unsigned char *buf=(const unsigned char *) sbuf, *ubuf;
  int length;

  const int Adpt[]={ 230, 230, 230, 230, 307, 409, 512, 614, 
		     768, 614, 512, 409, 307, 230, 230, 230 };
  const int AC1[]={ 256, 512, 0, 192, 240, 460, 392 };
  const int AC2[]={ 0, -256, 0, 64, 0, -208, -232 };

  int block_size=getWord(buf+32);
  int start_point=20+getDWord(buf+16);
  if (start_point+8 > size) goto ret_error;
  if (strncmp("data",sbuf+start_point,4)!=0) {
    start_point=getDWord(buf+start_point+4)+start_point+8;
  }
  if (strncmp("data",sbuf+start_point,4)!=0) goto ret_error;
  length=getDWord(buf+start_point+4);
  if (start_point+length>size) goto ret_error;
  start_point+=8;
  if (start_point>=size) goto ret_error;


  size=0;
  while(length>0) {
    int len=length>block_size?block_size:length;
    length-=len;
    ubuf=buf+start_point;
    start_point+=len;
    len-=7;

    int bp=ubuf[0];
    int delta=getSWord(ubuf+1);
    int s1=getSWord(ubuf+3);
    int s2=getSWord(ubuf+5);
    ubuf+=7;

    obuf[size++]=0xff & ((s2+32768)>>8);
    obuf[size++]=0xff & ((s1+32768)>>8);

    int coeff1=AC1[bp];
    int coeff2=AC2[bp];
    len*=2;
    int adrs=0;
    for(int i=0;i<len;i++){
      int val=(i&1)?(ubuf[adrs] & 15):(ubuf[adrs]&0xf0)>>4;
      if (val>7) val-=16;
      int x = ((s1*coeff1)+(s2*coeff2))/256+delta*val;
      if (x>32767) x=32767;
      if (x<-32768) x=-32768;
      s2=s1;
      s1=x;
      delta=Adpt[15 & val]*delta/256;
      if (delta<16) delta=16;
      
      obuf[size++]=0xff & ((x+32768)>>8);

      if ((i&1)!=0) adrs++;
    }
  }
  return obuf;

 ret_error:
  delete obuf;
  return NULL;
}

// EBXA
inline int GetHalf(int high_p, unsigned char *ubuf) {
  int val = (high_p == 0) ? *ubuf: (*ubuf)>>4;
  return (val & 7) - (val & 8);
}

inline void XADecode(char* &outp,char* sbuf,int& prev1,int& prev2){
  const int delta1[4]={0,0xf0,0x1cc,0x188}; // 꾮Ŭ˷׻....
  const int delta2[4]={0,0,-0xd0,-0xdc};
  int val;
  unsigned char* ubuf=reinterpret_cast<unsigned char*>(sbuf) + 4;
  unsigned char* ubuf2=reinterpret_cast<unsigned char*>(sbuf) + 16;
  for(int i=0;i<8;i++){
    int r=12-(ubuf[i]&15);
    int f=(ubuf[i]>>4)&15;
    for(int j=0;j<28;j++){
      val=(GetHalf((i&1),ubuf2+i/2+j*4)<<r)+
	((prev1*delta1[f]+prev2*delta2[f])>>8);
      if (val<-0x8000) val=-0x8000;
      if (val>0x7fff) val=0x7fff;
      *outp++=val & 255;
      *outp++=((0xff00 & val) >> 8)&255;
      prev2=prev1;
      prev1=val;
    }
  }
}

bool PlayXASound(const char* fname, int startLBA, int endLBA){
  FILE* soundfp=fopen(fname, "rb");
  if (soundfp==NULL) return false;
  int size=8*28*2*18*((endLBA-startLBA)>>4);
  int l=size+0x24+8;         // Data + Header Length
  fprintf(stderr,"size = %d\n",l);
  char buf0[128];
  char outbuf[l];
  char* outp=outbuf; 
  // Write WAV Header
  *outp++='R';  *outp++='I';  *outp++='F';  *outp++='F';

  *outp++=l & 255;
  *outp++=(l>>8) & 255;
  *outp++=(l>>16) & 255;
  *outp++=(l>>24) & 255;
  
  *outp++='W';  *outp++='A';  *outp++='V';  *outp++='E';
  *outp++='f';  *outp++='m';  *outp++='t';  *outp++=' ';
  *outp++=0x10; *outp++=0x00; *outp++=0x00; *outp++=0x00;
  *outp++=0x01; *outp++=0x00; *outp++=0x01; *outp++=0x00;
  *outp++=0xd4; *outp++=0x49; // Sample Rate 18900
  *outp++=0x00; *outp++=0x00;
  *outp++=0xa8; *outp++=0x93; // Sample Rate 18900*2
  *outp++=0x00; *outp++=0x00;
  *outp++=0x02; *outp++=0x00; *outp++=0x10; *outp++=0x00;
  *outp++='d';  *outp++='a';  *outp++='t';  *outp++='a';
  *outp++=size & 255;
  *outp++=(size>>8) & 255;
  *outp++=(size>>16) & 255;
  *outp++=(size>>24) & 255;

  int y=0,z=0;
  
  bool isWinFile=false;
  fseek(soundfp,0,SEEK_SET);
  if (fread(buf0,1,128,soundfp)==0) return false;
  if (memcmp(buf0,"RIFF",4)==0 && 
      memcmp(buf0+8,"CDXAfmt ",8)==0) isWinFile=true;
  if (isWinFile) fprintf(stderr,"WinFmt CDXA\n");
  for(;startLBA<endLBA;startLBA+=16){
    if (isWinFile){
      fseek(soundfp,startLBA*0x930+44+0x18,SEEK_SET);
    } else {
      fseek(soundfp,startLBA*0x900,SEEK_SET);
    }
    for(int i=0;i<18;i++){
      fread(buf0,1,128,soundfp);
      XADecode(outp,buf0,y,z);
    }
  }
  fclose(soundfp);
  fprintf(stderr,"Decoded size = %d\n",outp-outbuf);
  PlaySound(outbuf, outp-outbuf);
  return true;
}


#endif
